"""Functions to extract information from CasADi documentation.

Taken from
https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py.
"""

from collections import namedtuple
from contextlib import suppress
from itertools import dropwhile
from typing import Any, Callable
from warnings import warn

import casadi as cs


class _LambdaType:
    def __init__(self, func, typerepr):
        self.__typerepr = typerepr
        self.__func = func

    def __call__(self, val):
        return self.__func(val)

    def __repr__(self):
        return f"<type '{self.__typerepr}'>"

    def __str__(self):
        return repr(self)


_DocCell = namedtuple("_DocCell", ["id", "default", "doc"])
_TABLE_START = "+="  # string prefix that starts the table
_CELL_END = "+-"  # string prefix that ends the cell
_CELL_CONTENTS = "|"  # string prefix that continues the cell
_JOINS = ("", "", "", " ")  # how to join multiple lines in a given cell
_TYPES: dict[str, Callable] = {  # casadi types to python types
    "OT_INTEGER": int,
    "OT_STRING": str,
    "OT_REAL": float,
    "OT_INT": int,
    "OT_DICT": dict,
    "OT_DOUBLE": float,
    "OT_BOOL": bool,
    "OT_STR": str,
    "OT_INTVECTOR": _LambdaType(lambda x: [int(i) for i in x], "list[int]"),
    "OT_STRINGVECTOR": _LambdaType(lambda x: [str(i) for i in x], "list[str]"),
}


def _get_doc_cell(lines) -> _DocCell:
    """
    Returns a DocCell tuple for the set of lines.

    joins is a tuple of strings to say how to join multiple lines in a given
    cell. It must have exactly one entry for each cell
    """
    ncol = 4
    fields: tuple[list[str], ...] = ([], [], [], [])
    for line in lines:
        cells = line.split(" | ", ncol - 1)
        cells[0] = cells[0].lstrip().lstrip("|")
        cells[-1] = cells[-1].rstrip().rstrip("|")
        if len(cells) != ncol:
            raise ValueError("Wrong number of columns.")
        for i, c in enumerate(cells):
            fields[i].append(c.strip())

    id, type, default, doc = (j.join(f) for j, f in zip(_JOINS, fields))

    if typefunc := _TYPES.get(type):
        try:
            default = typefunc(default)
        except (ValueError, TypeError):
            includetype = True
            if default in ("None", "GenericType()"):
                default = None  # type: ignore[assignment]
            elif typefunc is int:
                with suppress(ValueError, TypeError):
                    default = int(float(default))  # type: ignore[assignment]
                    includetype = False
            if includetype:
                default = (default, typefunc)  # type: ignore[assignment]
    else:
        warn(f"Unknown type for '{id}', '{type}'.")
    return _DocCell(id, default, doc)


def _get_doc_dict(docstring: str) -> dict[str, tuple[Any, str]]:
    lineiter = dropwhile(
        lambda x: not x.startswith(_TABLE_START), docstring.split("\n")
    )
    try:
        next(lineiter)
    except StopIteration as e:
        raise ValueError("No table found!") from e
    thiscell: list[str] = []
    allcells: list[_DocCell] = []
    for line in lineiter:
        if line.startswith(_CELL_END):
            allcells.append(_get_doc_cell(thiscell))
            thiscell.clear()
        elif line.startswith(_CELL_CONTENTS):
            thiscell.append(line)
        else:
            break
    return {c.id: (c.default, c.doc) for c in allcells}


def get_casadi_plugins() -> dict[str, list[str]]:
    """Returns a dictionary of available casadi plugin as a dict of (type, name)."""

    func = getattr(cs, "CasadiMeta_getPlugins", getattr(cs, "CasadiMeta_plugins", None))
    if func is None:
        raise RuntimeError("Unable to get Casadi plugins.")

    all_plugins: str = func()
    plugins = (p.split("::") for p in all_plugins.split(";"))
    plugin_dict: dict[str, list[str]] = {}
    for group, name in plugins:
        if group not in plugin_dict:
            plugin_dict[group] = []
        plugin_dict[group].append(name)
    return plugin_dict


def list_available_solvers() -> dict[str, list[str]]:
    """Returns available solvers as a string or a dictionary."""
    availablesolvers = get_casadi_plugins()
    return {
        "nlp": availablesolvers.get("Nlpsol", []),
        "qp": availablesolvers.get("Conic", []) + availablesolvers.get("Qpsol", []),
    }


def get_solver_options(solver, display: bool = True) -> dict[str, tuple[Any, str]]:
    """Returns a dictionary of solver-specific options, with default value and
    description."""
    availablesolvers = list_available_solvers()
    if solver in availablesolvers["nlp"]:
        docstring = cs.doc_nlpsol(solver)
    elif solver in availablesolvers["qp"]:
        docstring = cs.doc_conic(solver)
    else:
        raise ValueError(f"Unknown solver: '{solver}'.")
    options: dict[str, tuple[Any, str]] = _get_doc_dict(docstring)
    if display:
        print("Available options [default] for %s:\n" % solver)
        for k in sorted(options.keys()):
            print(k, "[%r]: %s\n" % options[k])
    return options



__all__ = [
    "Mpc",
    "NlpScaling",
    "NlpSensitivity",
    "NonRetroactiveWrapper",
    "ScenarioBasedMpc",
    "Wrapper",
]

from .mpc.mpc import Mpc
from .mpc.scenario_based_mpc import ScenarioBasedMpc
from .scaling import NlpScaling
from .sensitivity import NlpSensitivity
from .wrapper import NonRetroactiveWrapper, Wrapper



from itertools import product
from typing import Union

import casadi as cs
import numpy as np

from .data import array2cs, cs2array


def hojacobian(ex: Union[cs.MX, cs.SX], x: Union[cs.MX, cs.SX]) -> np.ndarray:
    """Computes jacobian on higher-order matrices, i.e., with an output in 4D.

    Parameters
    ----------
    ex : casadi.MX, SX
        The expression to compute the jacobian for. Can be a matrix.
    x : casadi.MX, SX
        The variable to differentiate with respect to. Can be a matrix.

    Returns
    -------
    np.ndarray
        A 4D array of objects, where each entry `(i,j,k,m)` is the derivative
        of `ex[i,j]` w.r.t. `x[k,m]`.
    """
    return cs2array(cs.jacobian(cs.vec(ex), cs.vec(x))).reshape(
        ex.shape + x.shape, order="F"
    )


def hohessian(
    ex: Union[cs.MX, cs.SX], x: Union[cs.MX, cs.SX], y: Union[cs.MX, cs.SX, None] = None
) -> tuple[np.ndarray, np.ndarray]:
    """Computes hessian on higher-order matrices, i.e., with an output in 6D.

    Parameters
    ----------
    ex : casadi.MX, SX
        The expression to compute the hessin for. Can be a matrix.
    x : casadi.MX, SX
        The variable to differentiate with respect to. Can be a matrix.
    y : casadi.MX, SX, optional
        Use this argument to specify the second partial derivative, i.e., if
        the output requested is not the hessian of `ex` w.r.t. `x`, but rather
        the second-order partial derivatives of `ex` w.r.t. `x` and `y`.

    Returns
    -------
    np.ndarray, np.ndarray
        The first element is a 6D array of objects, where each entry
        `(i,j,k,m,l,p)` is the hessian of `ex[i,j]` w.r.t. `x[k,m], y[l,p]`,
        while the second element contains the jacobian instead (see
        `hojacobian` for details).
    """
    if y is None:
        y = x
    J = hojacobian(ex, x)
    H = np.empty(ex.shape + x.shape + y.shape, object)
    for i in product(*map(range, ex.shape)):
        H[i] = hojacobian(array2cs(J[i]), y)
    return H, J



__version__ = "1.5.10.post2"

__all__ = ["Nlp", "Solution", "multistart", "scaling"]

from . import multistart
from .core import scaling
from .core.solutions import Solution
from .nlps.nlp import Nlp



import pickle
import unittest
from itertools import product
from typing import TypeVar
from unittest.mock import Mock

import casadi as cs
import numpy as np
from parameterized import parameterized, parameterized_class

from csnlp.core.solutions import subsevalf
from csnlp.multistart import (
    MappedMultistartNlp,
    ParallelMultistartNlp,
    RandomStartPoint,
    RandomStartPoints,
    StackedMultistartNlp,
    StructuredStartPoint,
    StructuredStartPoints,
)
from csnlp.multistart.multistart_nlp import MultistartNlp, _find_best_sol, _n

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}
MULTI_NLP_CLASSES = [ParallelMultistartNlp, StackedMultistartNlp, MappedMultistartNlp]
TMultiNlp = TypeVar("TMultiNlp", *MULTI_NLP_CLASSES)


class TestStartPoints(unittest.TestCase):
    def test_random_start_points__raises__with_invalid_generator_method(self):
        method = "an_invalid_method"
        S = RandomStartPoints({"x": RandomStartPoint(method)}, 10)
        with self.assertRaisesRegex(
            AttributeError,
            f"'numpy.random._generator.Generator' object has no attribute '{method}'",
        ):
            list(S)

    def test_random_start_points__calls_random_method_with_args_and_kwargs(self):
        method = "uniform"
        multistarts = 10
        args, kwargs = object(), object()
        S = RandomStartPoints(
            {"x": RandomStartPoint(method, args, optional=kwargs)}, multistarts
        )
        S.np_random = Mock()
        points = list(S)
        self.assertEqual(len(points), multistarts)
        self.assertEqual(len(S.np_random.mock_calls), multistarts)
        getattr(S.np_random, method).assert_called_with(args, optional=kwargs)

    def test_random_start_points__returns_correct_values__when_seeded(self):
        multistarts = 5
        seed = 69
        biases = {"x": 0.5, "y": -0.2}
        scales = {"x": 0.1, "y": 0.3}
        S = RandomStartPoints(
            points={"x": RandomStartPoint("uniform"), "y": RandomStartPoint("normal")},
            multistarts=multistarts,
            biases=biases,
            scales=scales,
            seed=seed,
        )
        expecteds = [
            (0.5803723752156749, 0.33860174484688865),
            (0.8649616587869663, -0.9947886944731732),
            (0.47658484007382285, 1.0424725873226823),
            (0.5446347761155709, -0.5473146662455354),
            (0.008451104579143554, -0.34391256021075023),
        ]
        for actual, expected in zip(S, expecteds):
            self.assertAlmostEqual(actual["x"], expected[0] * scales["x"] + biases["x"])
            self.assertAlmostEqual(actual["y"], expected[1] * scales["y"] + biases["y"])

    def test_structured_start_points__returns_correct_values(self):
        multistarts = np.random.randint(10, 100)
        x_bnds = (np.random.rand() * 10, np.random.rand() * 100 + 20)
        y_bnds = (np.random.rand() * 10, np.random.rand() * 100 + 20)
        S = StructuredStartPoints(
            {"x": StructuredStartPoint(*x_bnds), "y": StructuredStartPoint(*y_bnds)},
            multistarts,
        )
        x_space = np.linspace(*x_bnds, multistarts)
        y_space = np.linspace(*y_bnds, multistarts)
        for actual, expected_x, expected_y in zip(S, x_space, y_space):
            self.assertEqual(actual["x"], expected_x)
            self.assertEqual(actual["y"], expected_y)


@parameterized_class("sym_type", [("SX",), ("MX",)])
class TestMultistartNlp(unittest.TestCase):
    def test_find_best_sol__returns_correct_solution(self):
        test_cases = [
            {
                "name": "All successful, varying 'f' values",
                "sols": [(1, True, ""), (2, True, ""), (3, True, "")],
                "expected": (1, True, ""),
            },
            {
                "name": "Mixed success, all feasible",
                "sols": [(1, False, ""), (2, True, ""), (3, True, "")],
                "expected": (2, True, ""),
            },
            {
                "name": "No successful, mixed feasibility",
                "sols": [(1, False, "infeasible"), (2, False, ""), (3, False, "")],
                "expected": (2, False, ""),
            },
            {
                "name": "All infeasible",
                "sols": [
                    (1, False, "infeasible"),
                    (2, False, "infeasible"),
                    (3, False, "infeasible"),
                ],
                "expected": (1, False, "infeasible"),
            },
            {
                "name": "Mixed success, infeasibility, varying 'f'",
                "sols": [
                    (2, True, ""),
                    (5, True, ""),
                    (3, False, ""),
                    (4, False, "infeasible"),
                ],
                "expected": (2, True, ""),
            },
        ]
        for case in test_cases:
            with self.subTest(name=case["name"]):
                sols = (
                    {"f": f, "stats": {"success": success, "return_status": status}}
                    for f, success, status in case["sols"]
                )
                actual = _find_best_sol(sols)
                self.assertTupleEqual(
                    (actual["f"], *actual["stats"].values()), case["expected"]
                )

    def test_init__raises__with_invalid_number_of_starts(self):
        with self.assertRaisesRegex(
            ValueError, "Number of scenarios must be positive and > 0."
        ):
            StackedMultistartNlp(starts=0, sym_type=self.sym_type)

    def test_variable_parameter_and_constraint__builds_correct_copies(self):
        N = 3
        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        y = nlp.variable("y", (3, 1))[0]
        z = nlp.variable("z")[0]
        p = nlp.parameter("p")
        nlp.constraint("c0", x**2, "<=", p)
        nlp.constraint("c1", y, "==", p)
        nlp.constraint("c2", z, ">=", -p)
        self.assertEqual(nlp.nx, nlp._stacked_nlp.nx // N)
        self.assertEqual(nlp.np, nlp._stacked_nlp.np // N)
        self.assertEqual(nlp.ng, nlp._stacked_nlp.ng // N)
        self.assertEqual(nlp.nh, nlp._stacked_nlp.nh // N)

    def test_minimize__sums_objectives_in_unique_function(self):
        N = 3
        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.minimize(cs.exp((x - 1) ** 2))
        x_ = cs.DM(np.random.randn(*x.shape))
        x_dict = {_n("x", i): x_ for i in range(N)}
        f1 = subsevalf(nlp.f, x, x_)
        f2 = subsevalf(nlp._stacked_nlp.f, nlp._stacked_nlp._vars, x_dict)
        self.assertAlmostEqual(f1, f2 / N)

    def test_solve__raises__with_both_flags_on(self):
        N = 3
        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
        with self.assertRaisesRegex(
            AssertionError,
            "`return_all_sols` and `return_stacked_sol` can't be both true.",
        ):
            nlp((None,), (None,), return_all_sols=True, return_stacked_sol=True)

    @parameterized.expand(product([False, True], MULTI_NLP_CLASSES))
    def test_solve__computes_right_solution(
        self, copy: bool, multinlp_cls: type[TMultiNlp]
    ):
        N = 3
        nlp = multinlp_cls(starts=N, sym_type=self.sym_type)
        x = nlp.variable("x", lb=-0.5, ub=1.4)[0]
        p = nlp.parameter("p")
        nlp.minimize(
            -0.3 * p * x**2
            - cs.exp(-10 * p * x**2)
            + cs.exp(-100 * p * (x - 1) ** 2)
            + cs.exp(-100 * p * (x - 1.5) ** 2)
        )
        nlp.init_solver(OPTS)
        if copy:
            nlp = nlp.copy()

        # solve manually
        x0s = [0.9, 0.5, 1.1]
        xfs, fs = [], []
        for x0 in x0s:
            sol = nlp.solve(pars={"p": 1.0}, vals0={"x": x0})
            xfs.append(sol.vals["x"])
            fs.append(sol.f)

        # solve with multistart
        args = ({"p": 1.0}, [{"x": x0} for x0 in x0s])
        best_sol = nlp.solve_multi(*args)
        all_sols = nlp.solve_multi(*args, return_all_sols=True)

        for xf, f, sol in zip(xfs, fs, all_sols):
            np.testing.assert_allclose(xf, sol.vals["x"])
            np.testing.assert_allclose(xf, sol.value(x))
            np.testing.assert_allclose(f, sol.f)
            np.testing.assert_allclose(f, sol.value(nlp.f))
        np.testing.assert_allclose(best_sol.f, min(fs))
        np.testing.assert_allclose(best_sol.value(nlp.f), min(fs))

    @parameterized.expand([(cls,) for cls in MULTI_NLP_CLASSES])
    def test_is_pickleable(self, multinlp_cls: type[TMultiNlp]):
        N = 3
        nlp = multinlp_cls(starts=N, sym_type=self.sym_type)
        x = nlp.variable("x", lb=-0.5, ub=1.4)[0]
        p = nlp.parameter("p")
        nlp.minimize(
            -0.3 * p * x**2
            - cs.exp(-10 * p * x**2)
            + cs.exp(-100 * p * (x - 1) ** 2)
            + cs.exp(-100 * p * (x - 1.5) ** 2)
        )
        nlp.init_solver(OPTS)
        nlp2: MultistartNlp = pickle.loads(pickle.dumps(nlp))
        self.assertEqual(nlp.name, nlp2.name)
        if multinlp_cls is StackedMultistartNlp:
            self.assertEqual(nlp._stacked_nlp.name, nlp2._stacked_nlp.name)


if __name__ == "__main__":
    unittest.main()
